@@ -16,7 +16,8 @@ import torch.nn.functional as F
 import warnings
 from joblib import Parallel, delayed
 
-from ._base import BaseModule, torchensemble_model_doc
+from ._base import BaseModule, BaseClassifier, BaseRegressor
+from ._base import torchensemble_model_doc
 from .utils import io
 from .utils import set_module
 from .utils import operator as op
@@ -216,7 +217,7 @@ class _BaseAdversarialTraining(BaseModule):
     """Implementation on the AdversarialTrainingClassifier.""",  # noqa: E501
     "model",
 )
-class AdversarialTrainingClassifier(_BaseAdversarialTraining):
+class AdversarialTrainingClassifier(_BaseAdversarialTraining, BaseClassifier):
     def __init__(self, **kwargs):
         super().__init__(**kwargs)
         self.is_classification = True
@@ -367,6 +368,12 @@ class AdversarialTrainingClassifier(_BaseAdversarialTraining):
                             " % | Historical Best: {:.3f} %"
                         )
                         self.logger.info(msg.format(epoch, acc, best_acc))
+                        if self.tb_logger:
+                            self.tb_logger.add_scalar(
+                                "adversarial_training/Validation_Acc",
+                                acc,
+                                epoch,
+                            )
 
                 # Update the scheduler
                 with warnings.catch_warnings():
@@ -383,32 +390,20 @@ class AdversarialTrainingClassifier(_BaseAdversarialTraining):
         if save_model and not test_loader:
             io.save(self, save_dir, self.logger)
 
-    @torchensemble_model_doc(
-        """Implementation on the evaluating stage of AdversarialTrainingClassifier.""",  # noqa: E501
-        "classifier_predict",
-    )
-    def predict(self, test_loader):
-        self.eval()
-        correct = 0
-        total = 0
+    @torchensemble_model_doc(item="classifier_evaluate")
+    def evaluate(self, test_loader, return_loss=False):
+        return super().evaluate(test_loader, return_loss)
 
-        for _, (data, target) in enumerate(test_loader):
-            data, target = data.to(self.device), target.to(self.device)
-            output = self.forward(data)
-            _, predicted = torch.max(output.data, 1)
-            correct += (predicted == target).sum().item()
-            total += target.size(0)
-
-        acc = 100 * correct / total
-
-        return acc
+    @torchensemble_model_doc(item="predict")
+    def predict(self, X, return_numpy=True):
+        return super().predict(X, return_numpy)
 
 
 @torchensemble_model_doc(
     """Implementation on the AdversarialTrainingRegressor.""",  # noqa: E501
     "model",
 )
-class AdversarialTrainingRegressor(_BaseAdversarialTraining):
+class AdversarialTrainingRegressor(_BaseAdversarialTraining, BaseRegressor):
     def __init__(self, **kwargs):
         super().__init__(**kwargs)
         self.is_classification = False
@@ -532,7 +527,7 @@ class AdversarialTrainingRegressor(_BaseAdversarialTraining):
                 if test_loader:
                     self.eval()
                     with torch.no_grad():
-                        mse = 0
+                        mse = 0.0
                         for _, (data, target) in enumerate(test_loader):
                             data = data.to(self.device)
                             target = target.to(self.device)
@@ -552,6 +547,12 @@ class AdversarialTrainingRegressor(_BaseAdversarialTraining):
                             " {:.5f} | Historical Best: {:.5f}"
                         )
                         self.logger.info(msg.format(epoch, mse, best_mse))
+                        if self.tb_logger:
+                            self.tb_logger.add_scalar(
+                                "adversirial_training/Validation_MSE",
+                                mse,
+                                epoch,
+                            )
 
                 # Update the scheduler
                 with warnings.catch_warnings():
@@ -565,18 +566,10 @@ class AdversarialTrainingRegressor(_BaseAdversarialTraining):
         if save_model and not test_loader:
             io.save(self, save_dir, self.logger)
 
-    @torchensemble_model_doc(
-        """Implementation on the evaluating stage of AdversarialTrainingRegressor.""",  # noqa: E501
-        "regressor_predict",
-    )
-    def predict(self, test_loader):
-        self.eval()
-        mse = 0
-        criterion = nn.MSELoss()
+    @torchensemble_model_doc(item="regressor_evaluate")
+    def evaluate(self, test_loader):
+        return super().evaluate(test_loader)
 
-        for batch_idx, (data, target) in enumerate(test_loader):
-            data, target = data.to(self.device), target.to(self.device)
-            output = self.forward(data)
-            mse += criterion(output, target)
-
-        return mse / len(test_loader)
+    @torchensemble_model_doc(item="predict")
+    def predict(self, X, return_numpy=True):
+        return super().predict(X, return_numpy)
