@@ -1,8 +1,8 @@
 """
   In bagging-based ensemble, each base estimator is trained independently.
   In addition, sampling with replacement is conducted on the training data
-  batch to encourage the diversity between different base estimators in the
-  ensemble.
+  batches to encourage the diversity between different base estimators in
+  the ensemble.
 """
 
 
@@ -13,7 +13,8 @@ import torch.nn.functional as F
 import warnings
 from joblib import Parallel, delayed
 
-from ._base import BaseModule, torchensemble_model_doc
+from ._base import BaseClassifier, BaseRegressor
+from ._base import torchensemble_model_doc
 from .utils import io
 from .utils import set_module
 from .utils import operator as op
@@ -95,13 +96,13 @@ def _parallel_fit_per_epoch(
 @torchensemble_model_doc(
     """Implementation on the BaggingClassifier.""", "model"
 )
-class BaggingClassifier(BaseModule):
+class BaggingClassifier(BaseClassifier):
     @torchensemble_model_doc(
         """Implementation on the data forwarding in BaggingClassifier.""",
         "classifier_forward",
     )
     def forward(self, x):
-        # Take the average over class distributions from all base estimators.
+        # Average over class distributions from all base estimators.
         outputs = [
             F.softmax(estimator(x), dim=1) for estimator in self.estimators_
         ]
@@ -239,6 +240,10 @@ class BaggingClassifier(BaseModule):
                             " % | Historical Best: {:.3f} %"
                         )
                         self.logger.info(msg.format(epoch, acc, best_acc))
+                        if self.tb_logger:
+                            self.tb_logger.add_scalar(
+                                "bagging/Validation_Acc", acc, epoch
+                            )
 
                 # Update the scheduler
                 with warnings.catch_warnings():
@@ -255,37 +260,25 @@ class BaggingClassifier(BaseModule):
         if save_model and not test_loader:
             io.save(self, save_dir, self.logger)
 
-    @torchensemble_model_doc(
-        """Implementation on the evaluating stage of BaggingClassifier.""",
-        "classifier_predict",
-    )
-    def predict(self, test_loader):
-        self.eval()
-        correct = 0
-        total = 0
+    @torchensemble_model_doc(item="classifier_evaluate")
+    def evaluate(self, test_loader, return_loss=False):
+        return super().evaluate(test_loader, return_loss)
 
-        for _, (data, target) in enumerate(test_loader):
-            data, target = data.to(self.device), target.to(self.device)
-            output = self.forward(data)
-            _, predicted = torch.max(output.data, 1)
-            correct += (predicted == target).sum().item()
-            total += target.size(0)
-
-        acc = 100 * correct / total
-
-        return acc
+    @torchensemble_model_doc(item="predict")
+    def predict(self, X, return_numpy=True):
+        return super().predict(X, return_numpy)
 
 
 @torchensemble_model_doc(
     """Implementation on the BaggingRegressor.""", "model"
 )
-class BaggingRegressor(BaseModule):
+class BaggingRegressor(BaseRegressor):
     @torchensemble_model_doc(
         """Implementation on the data forwarding in BaggingRegressor.""",
         "regressor_forward",
     )
     def forward(self, x):
-        # Take the average over predictions from all base estimators.
+        # Average over predictions from all base estimators.
         outputs = [estimator(x) for estimator in self.estimators_]
         pred = op.average(outputs)
 
@@ -396,7 +389,7 @@ class BaggingRegressor(BaseModule):
                 if test_loader:
                     self.eval()
                     with torch.no_grad():
-                        mse = 0
+                        mse = 0.0
                         for _, (data, target) in enumerate(test_loader):
                             data = data.to(self.device)
                             target = target.to(self.device)
@@ -416,6 +409,10 @@ class BaggingRegressor(BaseModule):
                             " {:.5f} | Historical Best: {:.5f}"
                         )
                         self.logger.info(msg.format(epoch, mse, best_mse))
+                        if self.tb_logger:
+                            self.tb_logger.add_scalar(
+                                "bagging/Validation_MSE", mse, epoch
+                            )
 
                 # Update the scheduler
                 with warnings.catch_warnings():
@@ -429,18 +426,10 @@ class BaggingRegressor(BaseModule):
         if save_model and not test_loader:
             io.save(self, save_dir, self.logger)
 
-    @torchensemble_model_doc(
-        """Implementation on the evaluating stage of BaggingRegressor.""",
-        "regressor_predict",
-    )
-    def predict(self, test_loader):
-        self.eval()
-        mse = 0
-        criterion = nn.MSELoss()
+    @torchensemble_model_doc(item="regressor_evaluate")
+    def evaluate(self, test_loader):
+        return super().evaluate(test_loader)
 
-        for batch_idx, (data, target) in enumerate(test_loader):
-            data, target = data.to(self.device), target.to(self.device)
-            output = self.forward(data)
-            mse += criterion(output, target)
-
-        return mse / len(test_loader)
+    @torchensemble_model_doc(item="predict")
+    def predict(self, X, return_numpy=True):
+        return super().predict(X, return_numpy)
