@@ -1,8 +1,8 @@
 """
   In fusion-based ensemble, predictions from all base estimators are
   first aggregated as an average output. After then, the training loss is
-  computed based on this average output and the ground-truth. The training loss
-  is then back-propagated to all base estimators simultaneously.
+  computed based on this average output and the ground-truth. The training
+  loss is then back-propagated to all base estimators simultaneously.
 """
 
 
@@ -10,7 +10,8 @@ import torch
 import torch.nn as nn
 import torch.nn.functional as F
 
-from ._base import BaseModule, torchensemble_model_doc
+from ._base import BaseClassifier, BaseRegressor
+from ._base import torchensemble_model_doc
 from .utils import io
 from .utils import set_module
 from .utils import operator as op
@@ -22,7 +23,7 @@ __all__ = ["FusionClassifier", "FusionRegressor"]
 @torchensemble_model_doc(
     """Implementation on the FusionClassifier.""", "model"
 )
-class FusionClassifier(BaseModule):
+class FusionClassifier(BaseClassifier):
     def _forward(self, x):
         """
         Implementation on the internal data forwarding in FusionClassifier.
@@ -91,6 +92,7 @@ class FusionClassifier(BaseModule):
         # Utils
         criterion = nn.CrossEntropyLoss()
         best_acc = 0.0
+        total_iters = 0
 
         # Training loop
         for epoch in range(epochs):
@@ -120,6 +122,11 @@ class FusionClassifier(BaseModule):
                                 epoch, batch_idx, loss, correct, data.size(0)
                             )
                         )
+                        if self.tb_logger:
+                            self.tb_logger.add_scalar(
+                                "fusion/Train_Loss", loss, total_iters
+                            )
+                total_iters += 1
 
             # Validation
             if test_loader:
@@ -146,6 +153,10 @@ class FusionClassifier(BaseModule):
                         " % | Historical Best: {:.3f} %"
                     )
                     self.logger.info(msg.format(epoch, acc, best_acc))
+                    if self.tb_logger:
+                        self.tb_logger.add_scalar(
+                            "fusion/Validation_Acc", acc, epoch
+                        )
 
             # Update the scheduler
             if hasattr(self, "scheduler_"):
@@ -154,34 +165,23 @@ class FusionClassifier(BaseModule):
         if save_model and not test_loader:
             io.save(self, save_dir, self.logger)
 
-    @torchensemble_model_doc(
-        """Implementation on the evaluating stage of FusionClassifier.""",
-        "classifier_predict",
-    )
-    def predict(self, test_loader):
-        self.eval()
-        correct = 0
-        total = 0
+    @torchensemble_model_doc(item="classifier_evaluate")
+    def evaluate(self, test_loader, return_loss=False):
+        return super().evaluate(test_loader, return_loss)
 
-        for _, (data, target) in enumerate(test_loader):
-            data, target = data.to(self.device), target.to(self.device)
-            output = self.forward(data)
-            _, predicted = torch.max(output.data, 1)
-            correct += (predicted == target).sum().item()
-            total += target.size(0)
-
-        acc = 100 * correct / total
-
-        return acc
+    @torchensemble_model_doc(item="predict")
+    def predict(self, X, return_numpy=True):
+        return super().predict(X, return_numpy)
 
 
 @torchensemble_model_doc("""Implementation on the FusionRegressor.""", "model")
-class FusionRegressor(BaseModule):
+class FusionRegressor(BaseRegressor):
     @torchensemble_model_doc(
         """Implementation on the data forwarding in FusionRegressor.""",
         "regressor_forward",
     )
     def forward(self, x):
+        # Average
         outputs = [estimator(x) for estimator in self.estimators_]
         pred = op.average(outputs)
 
@@ -234,6 +234,7 @@ class FusionRegressor(BaseModule):
         # Utils
         criterion = nn.MSELoss()
         best_mse = float("inf")
+        total_iters = 0
 
         # Training loop
         for epoch in range(epochs):
@@ -253,12 +254,17 @@ class FusionRegressor(BaseModule):
                     with torch.no_grad():
                         msg = "Epoch: {:03d} | Batch: {:03d} | Loss: {:.5f}"
                         self.logger.info(msg.format(epoch, batch_idx, loss))
+                        if self.tb_logger:
+                            self.tb_logger.add_scalar(
+                                "fusion/Train_Loss", loss, total_iters
+                            )
+                total_iters += 1
 
             # Validation
             if test_loader:
                 self.eval()
                 with torch.no_grad():
-                    mse = 0
+                    mse = 0.0
                     for _, (data, target) in enumerate(test_loader):
                         data = data.to(self.device)
                         target = target.to(self.device)
@@ -276,6 +282,10 @@ class FusionRegressor(BaseModule):
                         " Historical Best: {:.5f}"
                     )
                     self.logger.info(msg.format(epoch, mse, best_mse))
+                    if self.tb_logger:
+                        self.tb_logger.add_scalar(
+                            "fusion/Validation_MSE", mse, epoch
+                        )
 
             # Update the scheduler
             if hasattr(self, "scheduler_"):
@@ -284,18 +294,10 @@ class FusionRegressor(BaseModule):
         if save_model and not test_loader:
             io.save(self, save_dir, self.logger)
 
-    @torchensemble_model_doc(
-        """Implementation on the evaluating stage of FusionRegressor.""",
-        "regressor_predict",
-    )
-    def predict(self, test_loader):
-        self.eval()
-        mse = 0
-        criterion = nn.MSELoss()
+    @torchensemble_model_doc(item="regressor_evaluate")
+    def evaluate(self, test_loader):
+        return super().evaluate(test_loader)
 
-        for batch_idx, (data, target) in enumerate(test_loader):
-            data, target = data.to(self.device), target.to(self.device)
-            output = self.forward(data)
-            mse += criterion(output, target)
-
-        return mse / len(test_loader)
+    @torchensemble_model_doc(item="predict")
+    def predict(self, X, return_numpy=True):
+        return super().predict(X, return_numpy)
