@@ -10,7 +10,6 @@
 """
 
 
-import copy
 import math
 import torch
 import logging
@@ -19,10 +18,12 @@ import torch.nn as nn
 import torch.nn.functional as F
 from torch.optim.lr_scheduler import LambdaLR
 
-from ._base import BaseModule, torchensemble_model_doc
+from ._base import BaseModule, BaseClassifier, BaseRegressor
+from ._base import torchensemble_model_doc
 from .utils import io
 from .utils import set_module
 from .utils import operator as op
+from .utils.logging import get_tb_logger
 
 
 __all__ = [
@@ -109,6 +110,7 @@ class _BaseSnapshotEnsemble(BaseModule):
 
         self.device = torch.device("cuda" if cuda else "cpu")
         self.logger = logging.getLogger()
+        self.tb_logger = get_tb_logger()
 
         self.estimators_ = nn.ModuleList()
 
@@ -210,7 +212,7 @@ class _BaseSnapshotEnsemble(BaseModule):
 @torchensemble_model_doc(
     """Implementation on the SnapshotEnsembleClassifier.""", "seq_model"
 )
-class SnapshotEnsembleClassifier(_BaseSnapshotEnsemble):
+class SnapshotEnsembleClassifier(_BaseSnapshotEnsemble, BaseClassifier):
     def __init__(self, **kwargs):
         super().__init__(**kwargs)
         self.is_classification = True
@@ -251,12 +253,11 @@ class SnapshotEnsembleClassifier(_BaseSnapshotEnsemble):
             train_loader, self.is_classification
         )
 
-        # A dummy model used to generate snapshot ensembles
-        estimator_ = self._make_estimator()
+        estimator = self._make_estimator()
 
         # Set the optimizer and scheduler
         optimizer = set_module.set_optimizer(
-            estimator_, self.optimizer_name, **self.optimizer_args
+            estimator, self.optimizer_name, **self.optimizer_args
         )
 
         scheduler = self._set_scheduler(optimizer, epochs * len(train_loader))
@@ -265,10 +266,11 @@ class SnapshotEnsembleClassifier(_BaseSnapshotEnsemble):
         criterion = nn.CrossEntropyLoss()
         best_acc = 0.0
         counter = 0  # a counter on generating snapshots
+        total_iters = 0
         n_iters_per_estimator = epochs * len(train_loader) // self.n_estimators
 
         # Training loop
-        estimator_.train()
+        estimator.train()
         for epoch in range(epochs):
             for batch_idx, (data, target) in enumerate(train_loader):
 
@@ -279,7 +281,7 @@ class SnapshotEnsembleClassifier(_BaseSnapshotEnsemble):
                 optimizer = self._clip_lr(optimizer, lr_clip)
 
                 optimizer.zero_grad()
-                output = estimator_(data)
+                output = estimator(data)
                 loss = criterion(output, target)
                 loss.backward()
                 optimizer.step()
@@ -304,16 +306,26 @@ class SnapshotEnsembleClassifier(_BaseSnapshotEnsemble):
                                 batch_size,
                             )
                         )
+                        if self.tb_logger:
+                            self.tb_logger.add_scalar(
+                                "snapshot_ensemble/Train_Loss",
+                                loss,
+                                total_iters,
+                            )
+                        else:
+                            print("None")
 
                 # Snapshot ensemble updates the learning rate per iteration
                 # instead of per epoch.
                 scheduler.step()
                 counter += 1
+                total_iters += 1
 
             if counter % n_iters_per_estimator == 0:
 
                 # Generate and save the snapshot
-                snapshot = copy.deepcopy(estimator_)
+                snapshot = self._make_estimator()
+                snapshot.load_state_dict(estimator.state_dict())
                 self.estimators_.append(snapshot)
 
                 msg = "Save the snapshot model with index: {}"
@@ -346,35 +358,29 @@ class SnapshotEnsembleClassifier(_BaseSnapshotEnsemble):
                     self.logger.info(
                         msg.format(len(self.estimators_), acc, best_acc)
                     )
+                    if self.tb_logger:
+                        self.tb_logger.add_scalar(
+                            "snapshot_ensemble/Validation_Acc",
+                            acc,
+                            len(self.estimators_),
+                        )
 
         if save_model and not test_loader:
             io.save(self, save_dir, self.logger)
 
-    @torchensemble_model_doc(
-        """Implementation on the evaluating stage of SnapshotEnsembleClassifier.""",  # noqa: E501
-        "classifier_predict",
-    )
-    def predict(self, test_loader):
-        self.eval()
-        correct = 0
-        total = 0
+    @torchensemble_model_doc(item="classifier_evaluate")
+    def evaluate(self, test_loader, return_loss=False):
+        return super().evaluate(test_loader, return_loss)
 
-        for _, (data, target) in enumerate(test_loader):
-            data, target = data.to(self.device), target.to(self.device)
-            output = self.forward(data)
-            _, predicted = torch.max(output.data, 1)
-            correct += (predicted == target).sum().item()
-            total += target.size(0)
-
-        acc = 100 * correct / total
-
-        return acc
+    @torchensemble_model_doc(item="predict")
+    def predict(self, X, return_numpy=True):
+        return super().predict(X, return_numpy)
 
 
 @torchensemble_model_doc(
     """Implementation on the SnapshotEnsembleRegressor.""", "seq_model"
 )
-class SnapshotEnsembleRegressor(_BaseSnapshotEnsemble):
+class SnapshotEnsembleRegressor(_BaseSnapshotEnsemble, BaseRegressor):
     def __init__(self, **kwargs):
         super().__init__(**kwargs)
         self.is_classification = False
@@ -414,12 +420,11 @@ class SnapshotEnsembleRegressor(_BaseSnapshotEnsemble):
             train_loader, self.is_classification
         )
 
-        # A dummy model used to generate snapshot ensembles
-        estimator_ = self._make_estimator()
+        estimator = self._make_estimator()
 
         # Set the optimizer and scheduler
         optimizer = set_module.set_optimizer(
-            estimator_, self.optimizer_name, **self.optimizer_args
+            estimator, self.optimizer_name, **self.optimizer_args
         )
 
         scheduler = self._set_scheduler(optimizer, epochs * len(train_loader))
@@ -428,10 +433,11 @@ class SnapshotEnsembleRegressor(_BaseSnapshotEnsemble):
         criterion = nn.MSELoss()
         best_mse = float("inf")
         counter = 0  # a counter on generating snapshots
+        total_iters = 0
         n_iters_per_estimator = epochs * len(train_loader) // self.n_estimators
 
         # Training loop
-        estimator_.train()
+        estimator.train()
         for epoch in range(epochs):
             for batch_idx, (data, target) in enumerate(train_loader):
 
@@ -441,7 +447,7 @@ class SnapshotEnsembleRegressor(_BaseSnapshotEnsemble):
                 optimizer = self._clip_lr(optimizer, lr_clip)
 
                 optimizer.zero_grad()
-                output = estimator_(data)
+                output = estimator(data)
                 loss = criterion(output, target)
                 loss.backward()
                 optimizer.step()
@@ -461,15 +467,23 @@ class SnapshotEnsembleRegressor(_BaseSnapshotEnsemble):
                                 loss,
                             )
                         )
+                        if self.tb_logger:
+                            self.tb_logger.add_scalar(
+                                "snapshot_ensemble/Train_Loss",
+                                loss,
+                                total_iters,
+                            )
 
                 # Snapshot ensemble updates the learning rate per iteration
                 # instead of per epoch.
                 scheduler.step()
                 counter += 1
+                total_iters += 1
 
             if counter % n_iters_per_estimator == 0:
                 # Generate and save the snapshot
-                snapshot = copy.deepcopy(estimator_)
+                snapshot = self._make_estimator()
+                snapshot.load_state_dict(estimator.state_dict())
                 self.estimators_.append(snapshot)
 
                 msg = "Save the snapshot model with index: {}"
@@ -479,7 +493,7 @@ class SnapshotEnsembleRegressor(_BaseSnapshotEnsemble):
             if test_loader and counter % n_iters_per_estimator == 0:
                 self.eval()
                 with torch.no_grad():
-                    mse = 0
+                    mse = 0.0
                     for _, (data, target) in enumerate(test_loader):
                         data = data.to(self.device)
                         target = target.to(self.device)
@@ -499,22 +513,20 @@ class SnapshotEnsembleRegressor(_BaseSnapshotEnsemble):
                     self.logger.info(
                         msg.format(len(self.estimators_), mse, best_mse)
                     )
+                    if self.tb_logger:
+                        self.tb_logger.add_scalar(
+                            "snapshot_ensemble/Validation_MSE",
+                            mse,
+                            len(self.estimators_),
+                        )
 
         if save_model and not test_loader:
             io.save(self, save_dir, self.logger)
 
-    @torchensemble_model_doc(
-        """Implementation on the evaluating stage of SnapshotEnsembleRegressor.""",  # noqa: E501
-        "regressor_predict",
-    )
-    def predict(self, test_loader):
-        self.eval()
-        mse = 0
-        criterion = nn.MSELoss()
+    @torchensemble_model_doc(item="regressor_evaluate")
+    def evaluate(self, test_loader):
+        return super().evaluate(test_loader)
 
-        for batch_idx, (data, target) in enumerate(test_loader):
-            data, target = data.to(self.device), target.to(self.device)
-            output = self.forward(data)
-            mse += criterion(output, target)
-
-        return mse / len(test_loader)
+    @torchensemble_model_doc(item="predict")
+    def predict(self, X, return_numpy=True):
+        return super().predict(X, return_numpy)
